;;; CHICKEN binding to Mike's Arbitrary Precision Math Library.
;;; by Peter Wang <tjaden@users.sourceforge.net>
;;; Do whatever you like with this.

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(declare
 (foreign-declare "#include <m_apm.h>")
 (hide %make-apm %apm-free %apm-set-string! %apm-set-long! %apm-set-double!))

(define-foreign-type M_APM (c-pointer "M_APM_struct"))

; Internal procedures

(define (%make-apm)
  (let ((r (foreign-value "m_apm_init()" M_APM)))
    (set-finalizer! r %apm-free)
    r))

(define %apm-free (foreign-lambda void "m_apm_free" M_APM))

(define %apm-set-string! (foreign-lambda void "m_apm_set_string" M_APM nonnull-c-string))
(define %apm-set-long! (foreign-lambda void "m_apm_set_long" M_APM long))
(define %apm-set-double! (foreign-lambda void "m_apm_set_double" M_APM double))

; Helper macros

(define-macro (define-exported-foreign-variable name type string)
  (let ((tmpname (gensym)))
    `(begin
       (define-foreign-variable ,tmpname ,type ,string)
       (define ,name ,tmpname))))

(define-macro (with-r expr)		; captures `r'
  `(let ((r (%make-apm))) ,expr r))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

; Public stuff begins here

(define apm-precision (make-parameter 15))

;; Force constants to be initialised.
(foreign-code "m_apm_free(m_apm_init());")

(define-exported-foreign-variable mm-0 M_APM "MM_Zero")
(define-exported-foreign-variable mm-1 M_APM "MM_One")
(define-exported-foreign-variable mm-2 M_APM "MM_Two")
(define-exported-foreign-variable mm-3 M_APM "MM_Three")
(define-exported-foreign-variable mm-4 M_APM "MM_Four")
(define-exported-foreign-variable mm-5 M_APM "MM_Five")
(define-exported-foreign-variable mm-10 M_APM "MM_Ten")

(define-exported-foreign-variable mm-pi M_APM "MM_PI")
(define-exported-foreign-variable mm-pi/2 M_APM "MM_HALF_PI")
(define-exported-foreign-variable mm-2*pi M_APM "MM_2_PI")
(define-exported-foreign-variable mm-e M_APM "MM_E")

(define-exported-foreign-variable mm-log-e-base-10 M_APM "MM_LOG_E_BASE_10")
(define-exported-foreign-variable mm-log-10-base-e M_APM "MM_LOG_10_BASE_E")
(define-exported-foreign-variable mm-log-2-base-e M_APM "MM_LOG_2_BASE_E")
(define-exported-foreign-variable mm-log-3-base-e M_APM "MM_LOG_3_BASE_E")

;; This temporary buffer ought to be overkill for all real-life uses.
(declare (foreign-declare "static char _tmp_buf[1024];"))
(define-foreign-variable *_tmp_buf-threshold* int "(sizeof(_tmp_buf) - 16)")

(define mapm-lib-version
  (foreign-lambda*
   nonnull-c-string ()
   "return(m_apm_lib_version(_tmp_buf));"))

(define mapm-lib-short-version
  (foreign-lambda*
   nonnull-c-string ()
   "return(m_apm_lib_short_version(_tmp_buf));"))

(define (apm x)
  (ensure (lambda (x) (or (string? x) (number? x))) x
	  "(apm) bad argument type" x)
  (with-r (cond ((string? x) (%apm-set-string! r x))
		((fixnum? x) (%apm-set-long! r x))
		(else (%apm-set-double! r x)))))

;; These functions exist in MAPM, but I think they are probably a
;; little dangerous and unnecessary.
;(define apm-free-all-mem (foreign-lambda void "m_apm_free_all_mem"))
;(define apm-trim-mem-usage (foreign-lambda void "m_apm_trim_mem_usage"))

(define (apm->string x)
  ((if (< (apm-precision) *_tmp_buf-threshold*)
       (foreign-lambda*
	nonnull-c-string ((int n) (M_APM x))
	"m_apm_to_string(_tmp_buf, n, x); return(_tmp_buf);")
       (foreign-lambda*
	c-string* ((int n) (M_APM x))
	"char *s = malloc(n + 16);
         if (s) m_apm_to_string(s, n, x);
         return(s);"))
   (apm-precision) x))

(define (apm->fixpt-string x)
  ((if (< (apm-precision) *_tmp_buf-threshold*)
       (foreign-lambda*
	nonnull-c-string ((int n) (M_APM x))
	"m_apm_to_fixpt_string(_tmp_buf, n, x); return(_tmp_buf);")
       (foreign-lambda*
	c-string* ((int n) (M_APM x))
	"char *s = malloc(n + 16);
         if (s) m_apm_to_fixpt_string(s, n, x);
         return(s);"))
   (apm-precision) x))

;; m_apm_to_fixpt_stringex not required

(define (apm->fixpt-string-exp x radix sep sep-count)
  ((foreign-lambda*
    nonnull-c-string*
    ((int n) (M_APM x) (char radix) (char sep) (int sep_count))
    "return(m_apm_to_fixpt_stringexp(n, x, radix, sep, sep_count));")
   (apm-precision) x radix sep sep-count))

(define (apm->integer-string x)
  (let ((n (apm-significant-digits x)))
    (if (< n *_tmp_buf-threshold*)
	((foreign-lambda*
	  nonnull-c-string ((M_APM x))
	  "m_apm_to_integer_string(_tmp_buf, x); return(_tmp_buf);")
	 x)
	((foreign-lambda*
	  c-string* ((int n) (M_APM x))
	  "char *s = malloc(n + 16);
           if (s) m_apm_to_integer_string(s, x);
           return(s);")
	 n x))))

(define (apm-abs x)
  (with-r ((foreign-lambda void "m_apm_absolute_value" M_APM M_APM) r x)))

(define (apm-negate x)
  (with-r ((foreign-lambda void "m_apm_negate" M_APM M_APM) r x)))

;; Don't need this in scheme.
;(define (apm-copy x)
;  (with-r ((foreign-lambda void "m_apm_copy" M_APM M_APM) r x))) 

(define (apm-round n x)
  (with-r ((foreign-lambda void "m_apm_round" M_APM int M_APM) r n x)))

(define apm-compare
  (foreign-lambda int "m_apm_compare" M_APM M_APM))

(define apm-sign
  (foreign-lambda int "m_apm_sign" M_APM))

(define (apm-zero? x)
  (zero? (apm-sign x)))

(define (apm-positive? x)
  (= (apm-sign x) 1))

(define (apm-negative? x)
  (= (apm-sign x) -1))

(define (apm= x1 x2) (zero? (apm-compare x1 x2)))
(define (apm< x1 x2) (< (apm-compare x1 x2) 0))
(define (apm> x1 x2) (> (apm-compare x1 x2) 0))
(define (apm<= x1 x2) (<= (apm-compare x1 x2) 0))
(define (apm>= x1 x2) (>= (apm-compare x1 x2) 0))

(define apm-exponent
  (foreign-lambda int "m_apm_exponent" M_APM))

(define apm-significant-digits
  (foreign-lambda int "m_apm_significant_digits" M_APM))

(define apm-integer?
  (foreign-lambda bool "m_apm_is_integer" M_APM))

(define apm-even?
  (foreign-lambda bool "m_apm_is_even" M_APM))

(define apm-odd?
  (foreign-lambda bool "m_apm_is_odd" M_APM))

(define (apm-set-random-seed! x)
  ((foreign-lambda void "m_apm_set_random_seed" nonnull-c-string)
   (cond ((string? x) x)
	 ((number? x) (number->string x))
	 (else (error "apm-set-random-seed! expects a string or number" x)))))

(define (apm-get-random)
  (with-r ((foreign-lambda void "m_apm_get_random" M_APM) r)))

(define (apm+ x1 x2)
  (with-r ((foreign-lambda void "m_apm_add" M_APM M_APM M_APM) r x1 x2)))

(define (apm- x1 x2)
  (with-r ((foreign-lambda void "m_apm_subtract" M_APM M_APM M_APM) r x1 x2)))

(define (apm* x1 x2)
  (with-r ((foreign-lambda void "m_apm_multiply" M_APM M_APM M_APM) r x1 x2)))

(define (apm/ x1 x2)
  (with-r ((foreign-lambda void "m_apm_divide" M_APM int M_APM M_APM)
	   r (apm-precision) x1 x2)))

(define (apm-reciprocal x)
  (with-r ((foreign-lambda void "m_apm_reciprocal" M_APM int M_APM)
	   r (apm-precision) x)))

(define (apm-quotient x1 x2)
  (with-r ((foreign-lambda void "m_apm_integer_divide" M_APM M_APM M_APM)
	   r x1 x2)))

(define (apm-integer-div-rem x1 x2)
  (let ((quo (%make-apm)) (rem (%make-apm)))
    ((foreign-lambda void "m_apm_integer_div_rem" M_APM M_APM M_APM M_APM)
     quo rem x1 x2)
    (values quo rem)))

(define (apm-remainder x1 x2)
  (let ((quo (%make-apm)) (rem (%make-apm)))
    ((foreign-lambda void "m_apm_integer_div_rem" M_APM M_APM M_APM M_APM)
     quo rem x1 x2)
    rem))

(define (apm-factorial x)
  (with-r ((foreign-lambda void "m_apm_factorial" M_APM M_APM) r x)))

(define (apm-floor x)
  (with-r ((foreign-lambda void "m_apm_floor" M_APM M_APM) r x)))

(define (apm-ceiling x)
  (with-r ((foreign-lambda void "m_apm_ceil" M_APM M_APM) r x)))

(define (apm-gcd x1 x2)
  (with-r ((foreign-lambda void "m_apm_gcd" M_APM M_APM M_APM) r x1 x2)))

(define (apm-lcm x1 x2)
  (with-r ((foreign-lambda void "m_apm_lcm" M_APM M_APM M_APM) r x1 x2)))

(define (apm-sqrt x)
  (with-r ((foreign-lambda void "m_apm_sqrt" M_APM int M_APM)
	   r (apm-precision) x)))

(define (apm-cbrt x)			; cube root
  (with-r ((foreign-lambda void "m_apm_cbrt" M_APM int M_APM)
	   r (apm-precision) x)))

(define (apm-log x)
  (with-r ((foreign-lambda void "m_apm_log" M_APM int M_APM)
	   r (apm-precision) x)))

(define (apm-log10 x)
  (with-r ((foreign-lambda void "m_apm_log10" M_APM int M_APM)
	   r (apm-precision) x)))

(define (apm-exp x)
  (with-r ((foreign-lambda void "m_apm_exp" M_APM int M_APM)
	   r (apm-precision) x)))

(define (apm-expt x1 x2)
  (with-r ((foreign-lambda void "m_apm_pow" M_APM int M_APM M_APM)
	   r (apm-precision) x1 x2)))

(define (apm-integer-expt x i)
  (with-r ((foreign-lambda void "m_apm_integer_pow" M_APM int M_APM int)
	   r (apm-precision) x i)))

(define (apm-integer-expt/unrounded x i)
  (with-r ((foreign-lambda void "m_apm_integer_pow_nr" M_APM M_APM int)
	   r x i)))

(define (apm-sin x)
  (with-r ((foreign-lambda void "m_apm_sin" M_APM int M_APM)
	   r (apm-precision) x)))

(define (apm-cos x)
  (with-r ((foreign-lambda void "m_apm_cos" M_APM int M_APM)
	   r (apm-precision) x)))

(define (apm-sin-cos x)
  (let ((sv (%make-apm)) (cv (%make-apm)))
    ((foreign-lambda void "m_apm_sin_cos" M_APM M_APM int M_APM)
     sv cv (apm-precision) x)
    (values sv cv)))

(define (apm-tan x)
  (with-r ((foreign-lambda void "m_apm_tan" M_APM int M_APM)
	   r (apm-precision) x)))

(define (apm-asin x)
  (with-r ((foreign-lambda void "m_apm_arcsin" M_APM int M_APM)
	   r (apm-precision) x)))

(define (apm-acos x)
  (with-r ((foreign-lambda void "m_apm_arccos" M_APM int M_APM)
	   r (apm-precision) x)))

(define (apm-atan x)
  (with-r ((foreign-lambda void "m_apm_arctan" M_APM int M_APM)
	   r (apm-precision) x)))

(define (apm-atan2 x y)
  (with-r ((foreign-lambda void "m_apm_arctan2" M_APM int M_APM M_APM)
	   r (apm-precision) x y)))

(define (apm-sinh x)
  (with-r ((foreign-lambda void "m_apm_sinh" M_APM int M_APM)
	   r (apm-precision) x)))

(define (apm-cosh x)
  (with-r ((foreign-lambda void "m_apm_cosh" M_APM int M_APM)
	   r (apm-precision) x)))

(define (apm-tanh x)
  (with-r ((foreign-lambda void "m_apm_tanh" M_APM int M_APM)
	   r (apm-precision) x)))

(define (apm-asinh x)
  (with-r ((foreign-lambda void "m_apm_arcsinh" M_APM int M_APM)
	   r (apm-precision) x)))

(define (apm-acosh x)
  (with-r ((foreign-lambda void "m_apm_arccosh" M_APM int M_APM)
	   r (apm-precision) x)))

(define (apm-atanh x)
  (with-r ((foreign-lambda void "m_apm_arctanh" M_APM int M_APM)
	   r (apm-precision) x)))

;;; That's all folks.
